from yowsup.stacks import  YowStackBuilder
from .layer import YowsupCliLayer
from yowsup.layers import YowLayerEvent
from yowsup.layers.axolotl.props import PROP_IDENTITY_AUTOTRUST
import sys


class YowsupCliStack(object):
    def __init__(self, credentials):
        stackBuilder = YowStackBuilder()

        self._stack = stackBuilder\
            .pushDefaultLayers()\
            .push(YowsupCliLayer)\
            .build()

        self._stack.setCredentials(credentials)
        self._stack.setProp(PROP_IDENTITY_AUTOTRUST, True)

    def set_prop(self, prop, val):
        self._stack.setProp(prop, val)

    def start(self):
        print("Yowsup Cli client\n==================\nType /help for available commands\n")
        self._stack.broadcastEvent(YowLayerEvent(YowsupCliLayer.EVENT_START))

        try:
            self._stack.loop()
        except KeyboardInterrupt:
            print("\nYowsdown")
            sys.exit(0)



from yowsup.layers import YowLayer, YowLayerEvent
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.layers.auth import YowAuthenticationProtocolLayer
from yowsup.layers.protocol_media.protocolentities.iq_requestupload import RequestUploadIqProtocolEntity
from yowsup.layers.protocol_media.mediauploader import MediaUploader
from yowsup.layers.network.layer import YowNetworkLayer
from yowsup.layers.auth.protocolentities import StreamErrorProtocolEntity
from yowsup.layers import EventCallback
import inspect
import logging
logger = logging.getLogger(__name__)


class ProtocolEntityCallback(object):
    def __init__(self, entityType):
        self.entityType = entityType

    def __call__(self, fn):
        fn.entity_callback = self.entityType
        return fn


class YowInterfaceLayer(YowLayer):

    PROP_RECONNECT_ON_STREAM_ERR = "org.openwhatsapp.yowsup.prop.interface.reconnect_on_stream_error"

    def __init__(self):
        super(YowInterfaceLayer, self).__init__()
        self.reconnect = False
        self.entity_callbacks = {}
        self.iqRegistry = {}
        # self.receiptsRegistry = {}
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        for m in members:
            if hasattr(m[1], "entity_callback"):
                fname = m[0]
                fn = m[1]
                self.entity_callbacks[fn.entity_callback] = getattr(self, fname)

    def _sendIq(self, iqEntity, onSuccess=None, onError=None):
        assert iqEntity.getTag() == "iq", "Expected *IqProtocolEntity in _sendIq, got %s" % iqEntity.getTag()
        self.iqRegistry[iqEntity.getId()] = (iqEntity, onSuccess, onError)
        self.toLower(iqEntity)

    def processIqRegistry(self, entity):
        """
        :type entity: IqProtocolEntity
        """
        if entity.getTag() == "iq":
            iq_id = entity.getId()
            if iq_id in self.iqRegistry:
                originalIq, successClbk, errorClbk = self.iqRegistry[iq_id]
                del self.iqRegistry[iq_id]

                if entity.getType() == IqProtocolEntity.TYPE_RESULT and successClbk:
                    successClbk(entity, originalIq)
                elif entity.getType() == IqProtocolEntity.TYPE_ERROR and errorClbk:
                    errorClbk(entity, originalIq)
                return True

        return False

    def getOwnJid(self, full=True):
        return self.getLayerInterface(YowAuthenticationProtocolLayer).getUsername(full)

    def connect(self):
        self.getLayerInterface(YowNetworkLayer).connect()

    def disconnect(self):
        disconnectEvent = YowLayerEvent(YowNetworkLayer.EVENT_STATE_DISCONNECT)
        self.broadcastEvent(disconnectEvent)

    def send(self, data):
        self.toLower(data)

    def receive(self, entity):
        if not self.processIqRegistry(entity):
            entityType = entity.getTag()
            if entityType in self.entity_callbacks:
                self.entity_callbacks[entityType](entity)
            else:
                self.toUpper(entity)

    @ProtocolEntityCallback("stream:error")
    def onStreamError(self, streamErrorEntity):
        logger.error(streamErrorEntity)
        if self.getProp(self.__class__.PROP_RECONNECT_ON_STREAM_ERR, True):
            if streamErrorEntity.getErrorType() == StreamErrorProtocolEntity.TYPE_CONFLICT:
                logger.warn("Not reconnecting because you signed in in another location")
            else:
                logger.info("Initiating reconnect")
                self.reconnect = True
        else:
            logger.warn("Not reconnecting because property %s is not set" %
                        self.__class__.PROP_RECONNECT_ON_STREAM_ERR)
        self.toUpper(streamErrorEntity)
        self.disconnect()

    @EventCallback(YowNetworkLayer.EVENT_STATE_CONNECTED)
    def onConnected(self, yowLayerEvent):
        self.reconnect = False

    @EventCallback(YowNetworkLayer.EVENT_STATE_DISCONNECTED)
    def onDisconnected(self, yowLayerEvent):
        if self.reconnect:
            self.reconnect = False
            self.connect()

    def _sendMediaMessage(self, builder, success, error=None, progress=None):
        # axolotlIface = self.getLayerInterface(YowAxolotlLayer)
        # if axolotlIface:
        #     axolotlIface.encryptMedia(builder)

        iq = RequestUploadIqProtocolEntity(
            builder.mediaType, filePath=builder.getFilepath(), encrypted=builder.isEncrypted())

        def successFn(resultEntity, requestUploadEntity): return self.__onRequestUploadSuccess(
            resultEntity, requestUploadEntity, builder, success, error, progress)

        def errorFn(errorEntity, requestUploadEntity): return self.__onRequestUploadError(
            errorEntity, requestUploadEntity, error)
        self._sendIq(iq, successFn, errorFn)

    def __onRequestUploadSuccess(self, resultRequestUploadIqProtocolEntity, requestUploadEntity, builder, success, error=None, progress=None):
        if(resultRequestUploadIqProtocolEntity.isDuplicate()):
            return success(builder.build(resultRequestUploadIqProtocolEntity.getUrl(), resultRequestUploadIqProtocolEntity.getIp()))
        else:
            def successFn(path, jid, url): return self.__onMediaUploadSuccess(
                builder, url, resultRequestUploadIqProtocolEntity.getIp(), success)

            def errorFn(path, jid, errorText): return self.__onMediaUploadError(
                builder, errorText, error)

            mediaUploader = MediaUploader(builder.jid, self.getOwnJid(), builder.getFilepath(),
                                          resultRequestUploadIqProtocolEntity.getUrl(),
                                          resultRequestUploadIqProtocolEntity.getResumeOffset(),
                                          successFn, errorFn, progress, asynchronous=True)
            mediaUploader.start()

    def __onRequestUploadError(self, errorEntity, requestUploadEntity, builder, error=None):
        if error:
            return error(errorEntity.code, errorEntity.text, errorEntity.backoff)

    def __onMediaUploadSuccess(self, builder, url, ip, successClbk):
        messageNode = builder.build(url, ip)
        return successClbk(messageNode)

    def __onMediaUploadError(self, builder, errorText, errorClbk=None):
        if errorClbk:
            return errorClbk(0, errorText, 0)

    def __str__(self):
        return "Interface Layer"



from yowsup.common import YowConstants
from yowsup.layers import YowLayerEvent, YowProtocolLayer, EventCallback
from yowsup.layers.network import YowNetworkLayer
from .protocolentities import *
from .layer_interface_authentication import YowAuthenticationProtocolLayerInterface
from .protocolentities import StreamErrorProtocolEntity


class YowAuthenticationProtocolLayer(YowProtocolLayer):
    EVENT_AUTHED  = "org.openwhatsapp.yowsup.event.auth.authed"
    EVENT_AUTH = "org.openwhatsapp.yowsup.event.auth"
    PROP_CREDENTIALS = "org.openwhatsapp.yowsup.prop.auth.credentials"
    PROP_PASSIVE = "org.openwhatsapp.yowsup.prop.auth.passive"

    def __init__(self):
        handleMap = {
            "stream:features": (self.handleStreamFeatures, None),
            "failure": (self.handleFailure, None),
            "success": (self.handleSuccess, None),
            "stream:error": (self.handleStreamError, None),
        }
        super(YowAuthenticationProtocolLayer, self).__init__(handleMap)
        self.interface = YowAuthenticationProtocolLayerInterface(self)
        self.credentials = None #left for backwards-compat
        self._credentials = None #new style set

    def __str__(self):
        return "Authentication Layer"

    @EventCallback(YowNetworkLayer.EVENT_STATE_CONNECTED)
    def on_connected(self, event):
        self.broadcastEvent(
            YowLayerEvent(
                self.EVENT_AUTH,
                credentials=self.getProp(self.PROP_CREDENTIALS),
                passive=self.getProp(self.PROP_PASSIVE, False)
            )
        )

    def setCredentials(self, credentials):
        self.setProp(self.PROP_CREDENTIALS, credentials) #keep for now
        self._credentials = credentials

#
    def getUsername(self, full = False):
        if self._credentials:
            return self._credentials[0] if not full else ("%s@%s" % (self._credentials[0], YowConstants.WHATSAPP_SERVER))
        else:
            prop = self.getProp(YowAuthenticationProtocolLayer.PROP_CREDENTIALS)
            return prop[0] if prop else None

    def handleStreamFeatures(self, node):
        nodeEntity = StreamFeaturesProtocolEntity.fromProtocolTreeNode(node)
        self.toUpper(nodeEntity)

    def handleSuccess(self, node):
        successEvent = YowLayerEvent(self.__class__.EVENT_AUTHED, passive = self.getProp(self.__class__.PROP_PASSIVE))
        self.broadcastEvent(successEvent)
        nodeEntity = SuccessProtocolEntity.fromProtocolTreeNode(node)
        self.toUpper(nodeEntity)

    def handleFailure(self, node):
        nodeEntity = FailureProtocolEntity.fromProtocolTreeNode(node)
        self.toUpper(nodeEntity)
        self.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_DISCONNECT, reason = "Authentication Failure"))

    def handleStreamError(self, node):
        nodeEntity = StreamErrorProtocolEntity.fromProtocolTreeNode(node)
        errorType = nodeEntity.getErrorType()

        if not errorType:
            raise NotImplementedError("Unhandled stream:error node:\n%s" % node)

        self.toUpper(nodeEntity)



from yowsup.layers.protocol_messages.proto.e2e_pb2 import Message
from yowsup.layers.axolotl.protocolentities import *
from yowsup.layers.auth.layer_authentication import YowAuthenticationProtocolLayer
from yowsup.layers.protocol_groups.protocolentities import InfoGroupsIqProtocolEntity, InfoGroupsResultIqProtocolEntity
from axolotl.protocol.whispermessage import WhisperMessage
from yowsup.layers.protocol_messages.protocolentities.message import MessageAttributes
from .layer_base import AxolotlBaseLayer

import logging

logger = logging.getLogger(__name__)


class AxolotlSendLayer(AxolotlBaseLayer):
    MAX_SENT_QUEUE = 100

    def __init__(self):
        super(AxolotlSendLayer, self).__init__()

        self.sessionCiphers = {}
        self.groupCiphers = {}
        '''
            Sent messages will be put in Queue until we receive a receipt for them.
            This is for handling retry receipts which requires re-encrypting and resend of the original message
            As the receipt for a sent message might arrive at a different yowsup instance,
            ideally the original message should be fetched from a persistent storage.
            Therefore, if the original message is not in sentQueue for any reason, we will
            notify the upper layers and let them handle it.
        '''
        self.sentQueue = []

    def __str__(self):
        return "Axolotl Layer"

    def send(self, node):
        if node.tag == "message" and node["to"] not in self.skipEncJids:
            self.processPlaintextNodeAndSend(node)
        else:
            self.toLower(node)

    def receive(self, protocolTreeNode):
        if not self.processIqRegistry(protocolTreeNode):
            if protocolTreeNode.tag == "receipt":
                '''
                Going to keep all group message enqueued, as we get receipts from each participant
                So can't just remove it on first receipt. Therefore, the MAX queue length mechanism should better be working
                '''
                messageNode = self.getEnqueuedMessageNode(protocolTreeNode["id"], protocolTreeNode["participant"] is not None)
                if not messageNode:
                    logger.debug("Axolotl layer does not have the message, bubbling it upwards")
                    self.toUpper(protocolTreeNode)
                elif protocolTreeNode["type"] == "retry":
                    logger.info("Got retry to for message %s, and Axolotl layer has the message" % protocolTreeNode["id"])
                    retryReceiptEntity = RetryIncomingReceiptProtocolEntity.fromProtocolTreeNode(protocolTreeNode)
                    self.toLower(retryReceiptEntity.ack().toProtocolTreeNode())
                    self.getKeysFor(
                        [protocolTreeNode["participant"] or protocolTreeNode["from"]],
                        lambda successJids, b: self.processPlaintextNodeAndSend(messageNode, retryReceiptEntity) if len(successJids) == 1 else None
                    )
                else:
                    #not interested in any non retry receipts, bubble upwards
                    self.toUpper(protocolTreeNode)

    def processPlaintextNodeAndSend(self, node, retryReceiptEntity = None):
        recipient_id = node["to"].split('@')[0]
        isGroup = "-" in recipient_id

        if isGroup:
            self.sendToGroup(node, retryReceiptEntity)
        elif self.manager.session_exists(recipient_id):
            self.sendToContact(node)
        else:
            self.getKeysFor([node["to"]], lambda successJids, b: self.sendToContact(node) if len(successJids) == 1 else self.toLower(node), lambda: self.toLower(node))

    def enqueueSent(self, node):
        logger.debug("enqueueSent(node=[omitted])")
        if len(self.sentQueue) >= self.__class__.MAX_SENT_QUEUE:
            logger.warn("Discarding queued node without receipt")
            self.sentQueue.pop(0)
        self.sentQueue.append(node)

    def getEnqueuedMessageNode(self, messageId, keepEnqueued = False):
        for i in range(0, len(self.sentQueue)):
            if self.sentQueue[i]["id"] == messageId:
                if keepEnqueued:
                    return self.sentQueue[i]
                return self.sentQueue.pop(i)

    def sendEncEntities(self, node, encEntities, participant=None):
        logger.debug("sendEncEntities(node=[omitted], encEntities=[omitted], participant=%s)" % participant)
        message_attrs = MessageAttributes.from_message_protocoltreenode(node)
        message_attrs.participant = participant
        messageEntity = EncryptedMessageProtocolEntity(
            encEntities,
            node["type"],
            message_attrs
        )
        # if participant is set, this message is directed to that specific participant as a result of a retry, therefore
        # we already have the original group message and there is no need to store it again.
        if participant is None:
            self.enqueueSent(node)
        self.toLower(messageEntity.toProtocolTreeNode())

    def sendToContact(self, node):
        recipient_id = node["to"].split('@')[0]

        protoNode = node.getChild("proto")
        messageData = protoNode.getData()

        ciphertext = self.manager.encrypt(
            recipient_id,
            messageData
        )
        mediaType = protoNode["mediatype"]
        return self.sendEncEntities(node, [EncProtocolEntity(EncProtocolEntity.TYPE_MSG if ciphertext.__class__ == WhisperMessage else EncProtocolEntity.TYPE_PKMSG, 2, ciphertext.serialize(), mediaType)])

    def sendToGroupWithSessions(self, node, jidsNeedSenderKey = None, retryCount=0):
        """
        For each jid in jidsNeedSenderKey will create a pkmsg enc node with the associated jid.
        If retryCount > 0 and we have only one jidsNeedSenderKey, this is a retry requested by a specific participant
        and this message is to be directed at specific at that participant indicated by jidsNeedSenderKey[0]. In this
        case the participant's jid would go in the parent's EncryptedMessage and not into the enc node.
        """
        logger.debug(
            "sendToGroupWithSessions(node=[omitted], jidsNeedSenderKey=%s, retryCount=%d)" % (jidsNeedSenderKey, retryCount)
        )
        jidsNeedSenderKey = jidsNeedSenderKey or []
        groupJid = node["to"]
        protoNode = node.getChild("proto")
        encEntities = []
        participant = jidsNeedSenderKey[0] if len(jidsNeedSenderKey) == 1 and retryCount > 0 else None
        if len(jidsNeedSenderKey):
            senderKeyDistributionMessage = self.manager.group_create_skmsg(groupJid)
            for jid in jidsNeedSenderKey:
                message =  self.serializeSenderKeyDistributionMessageToProtobuf(node["to"], senderKeyDistributionMessage)
                if retryCount > 0:
                    message.MergeFromString(protoNode.getData())
                ciphertext = self.manager.encrypt(jid.split('@')[0], message.SerializeToString())
                encEntities.append(
                    EncProtocolEntity(
                            EncProtocolEntity.TYPE_MSG if ciphertext.__class__ == WhisperMessage else EncProtocolEntity.TYPE_PKMSG
                        , 2, ciphertext.serialize(), protoNode["mediatype"],  jid=None if participant else jid
                    )
                )

        if not retryCount:
            messageData = protoNode.getData()
            ciphertext = self.manager.group_encrypt(groupJid, messageData)
            mediaType = protoNode["mediatype"]

            encEntities.append(EncProtocolEntity(EncProtocolEntity.TYPE_SKMSG, 2, ciphertext, mediaType))

        self.sendEncEntities(node, encEntities, participant)

    def ensureSessionsAndSendToGroup(self, node, jids):
        logger.debug("ensureSessionsAndSendToGroup(node=[omitted], jids=%s)" % jids)
        jidsNoSession = []
        for jid in jids:
            if not self.manager.session_exists(jid.split('@')[0]):
                jidsNoSession.append(jid)

        if len(jidsNoSession):
            self.getKeysFor(jidsNoSession, lambda successJids, b: self.sendToGroupWithSessions(node, successJids))
        else:
            self.sendToGroupWithSessions(node, jids)

    def sendToGroup(self, node, retryReceiptEntity = None):
        """
        Group send sequence:
        check if senderkeyrecord exists
            no: - create,
                - get group jids from info request
                - for each jid without a session, get keys to create the session
                - send message with dist key for all participants
            yes:
                - send skmsg without any dist key

        received retry for a participant
            - request participants keys
            - send message with dist key only + conversation, only for this participat
        """
        logger.debug("sendToGroup(node=[omitted], retryReceiptEntity=[%s])" %
                     ("[retry_count=%s, retry_jid=%s]" % (
                         retryReceiptEntity.getRetryCount(), retryReceiptEntity.getRetryJid())
                      ) if retryReceiptEntity is not None else None)

        groupJid = node["to"]
        ownJid = self.getLayerInterface(YowAuthenticationProtocolLayer).getUsername(True)

        senderKeyRecord = self.manager.load_senderkey(node["to"])

        def sendToGroup(resultNode, requestEntity):
            groupInfo = InfoGroupsResultIqProtocolEntity.fromProtocolTreeNode(resultNode)
            jids = list(groupInfo.getParticipants().keys()) #keys in py3 returns dict_keys
            if ownJid in jids:
                jids.remove(ownJid)
            return self.ensureSessionsAndSendToGroup(node, jids)

        if senderKeyRecord.isEmpty():
            logger.debug("senderKeyRecord is empty, requesting group info")
            groupInfoIq = InfoGroupsIqProtocolEntity(groupJid)
            self._sendIq(groupInfoIq, sendToGroup)
        else:
            logger.debug("We have a senderKeyRecord")
            retryCount = 0
            jidsNeedSenderKey = []
            if retryReceiptEntity is not None:
                retryCount = retryReceiptEntity.getRetryCount()
                jidsNeedSenderKey.append(retryReceiptEntity.getRetryJid())
            self.sendToGroupWithSessions(node, jidsNeedSenderKey, retryCount)

    def serializeSenderKeyDistributionMessageToProtobuf(self, groupId, senderKeyDistributionMessage, message = None):
        m = message or Message()
        m.sender_key_distribution_message.group_id = groupId
        m.sender_key_distribution_message.axolotl_sender_key_distribution_message = senderKeyDistributionMessage.serialize()
        m.sender_key_distribution_message.axolotl_sender_key_distribution_message = senderKeyDistributionMessage.serialize()
        # m.conversation = text
        return m



from yowsup.layers.protocol_messages.protocolentities.message_text import TextMessageProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_messages.protocolentities.test_message import MessageProtocolEntityTest
from yowsup.layers.protocol_messages.proto.e2e_pb2 import Message


class TextMessageProtocolEntityTest(MessageProtocolEntityTest):
    def setUp(self):
        super(TextMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = TextMessageProtocolEntity
        m = Message()
        m.conversation = "body_data"
        proto_node = ProtocolTreeNode("proto", {}, None, m.SerializeToString())
        self.node.addChild(proto_node)



class ContextInfo(object):
    def __init__(self,
                 stanza_id=None,
                 participant=None,
                 quoted_message=None,
                 remote_jid=None,
                 mentioned_jid=None,
                 edit_version=None,
                 revoke_message=None
                 ):
        self._stanza_id= stanza_id
        self._participant = participant
        self._quoted_message = quoted_message
        self._remote_jid = remote_jid
        self._mentioned_jid = mentioned_jid
        self._edit_version = edit_version
        self._revoke_message = revoke_message

    @property
    def stanza_id(self):
        return self._stanza_id

    @property
    def participant(self):
        return self._participant

    @property
    def quoted_message(self):
        return self._quoted_message

    @property
    def remote_jid(self):
        return self._remote_jid

    @property
    def mentioned_jid(self):
        return self._mentioned_jid

    @property
    def edit_version(self):
        return self._edit_version

    @property
    def revoke_message(self):
        return self._revoke_message



from yowsup.config.base.transform import ConfigTransform
import types


class PropsTransform(ConfigTransform):
    def __init__(self, transform_map=None, reverse_map=None):
        self._transform_map = transform_map or {}
        self._reverse_map = reverse_map or {}

    def transform(self, data):
        """
        :param data:
        :type data: dict
        :return:
        :rtype: dict
        """
        out = {}
        for key, val in data.items():
            if key in self._transform_map:
                target = self._transform_map[key]
                key, val = target(key, val) if type(target) == types.FunctionType else (key, target)

            out[key] = val


        return out

    def reverse(self, data):
        transformed_dict = {}

        for key, val in data.items():
            if key in self._reverse_map:
                target = self._reverse_map[key]
                key, val = target(key, val) if type(target) == types.FunctionType else (key, target)

            transformed_dict[key] = val

        return transformed_dict



from yowsup.stacks import  YowStackBuilder
from .layer import SendLayer
from yowsup.layers import YowLayerEvent
from yowsup.layers.auth import YowAuthenticationProtocolLayer
from yowsup.layers.network import YowNetworkLayer


class YowsupSendStack(object):
    def __init__(self, credentials, messages):
        """
        :param credentials:
        :param messages: list of (jid, message) tuples
        :return:
        """
        stackBuilder = YowStackBuilder()

        self._stack = stackBuilder\
            .pushDefaultLayers()\
            .push(SendLayer)\
            .build()

        self._stack.setProp(SendLayer.PROP_MESSAGES, messages)
        self._stack.setProp(YowAuthenticationProtocolLayer.PROP_PASSIVE, True)
        self._stack.setCredentials(credentials)

    def set_prop(self, key, val):
        self._stack.setProp(key, val)

    def start(self):
        self._stack.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECT))
        self._stack.loop()



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class StreamErrorProtocolEntity(ProtocolEntity):
    TYPE_CONFLICT = "conflict"
    '''
     <stream:error>
        <conflict></conflict>
        <text>Replaced by new connection</text>
     </stream:error>
    '''

    TYPE_ACK = "ack"
    '''
     <stream:error>
        <ack></ack>
     </stream:error>
    '''

    TYPE_XML_NOT_WELL_FORMED = "xml-not-well-formed"
    '''
    <stream:error>
        <xml-not-well-formed>
        </xml-not-well-formed>
    </stream:error>
    '''

    TYPES = (TYPE_CONFLICT, TYPE_ACK, TYPE_XML_NOT_WELL_FORMED)

    def __init__(self, data = None):
        super(StreamErrorProtocolEntity, self).__init__("stream:error")
        data = data or {}
        self.setErrorData(data)

    def setErrorData(self, data):
        self.data = data

    def getErrorData(self):
        return self.data

    def getErrorType(self):
        for k in self.data.keys():
            if k in self.__class__.TYPES:
                return k

    def __str__(self):
        out  = "Stream Error type: %s\n" % self.getErrorType()
        out += "%s" % self.getErrorData()
        out += "\n"

        return out

    def toProtocolTreeNode(self):
        node = super(StreamErrorProtocolEntity, self).toProtocolTreeNode()
        type = self.getErrorType()
        node.addChild(ProtocolTreeNode(type))
        if type == self.__class__.TYPE_CONFLICT and "text" in self.data:
            node.addChild(ProtocolTreeNode("text", data=self.data["text"]))

        return node


    @staticmethod
    def fromProtocolTreeNode(protocolTreeNode):
        data = {}
        for child in protocolTreeNode.getAllChildren():
            data[child.tag] = child.data
        return StreamErrorProtocolEntity(data)



from yowsup.layers.protocol_media.mediacipher import MediaCipher
from yowsup.layers.protocol_media.protocolentities \
    import ImageDownloadableMediaMessageProtocolEntity, AudioDownloadableMediaMessageProtocolEntity,\
    VideoDownloadableMediaMessageProtocolEntity, DocumentDownloadableMediaMessageProtocolEntity, \
    ContactMediaMessageProtocolEntity, DownloadableMediaMessageProtocolEntity

import threading
from tqdm import tqdm
import requests
import logging
import math
import sys
import os
import base64

if sys.version_info >= (3, 0):
    from queue import Queue
else:
    from Queue import Queue

logger = logging.getLogger(__name__)


class SinkWorker(threading.Thread):
    def __init__(self, storage_dir):
        super(SinkWorker, self).__init__()
        self.daemon = True
        self._storage_dir = storage_dir
        self._jobs = Queue()
        self._media_cipher = MediaCipher()

    def enqueue(self, media_message_protocolentity):
        self._jobs.put(media_message_protocolentity)

    def _create_progress_iterator(self, iterable, niterations, desc):
        return tqdm(
                    iterable, total=niterations, unit='KB', dynamic_ncols=True,
                    unit_scale=True, leave=True, desc=desc, ascii=True)

    def _download(self, url):
        response = requests.get(url, stream=True)
        total_size = int(response.headers.get('content-length', 0))
        logger.debug("%s total size is %s, downloading" % (url, total_size))
        block_size = 1024
        wrote = 0
        enc_data = b""

        for data in self._create_progress_iterator(
                response.iter_content(block_size), math.ceil(total_size // block_size) + 1, "Download       ",
        ):
            wrote = wrote + len(data)
            enc_data = enc_data + data

        if total_size != 0 and wrote != total_size:
            logger.error("Something went wrong")
            return None

        return enc_data

    def _decrypt(self, ciphertext, ref_key, media_info):
        length_kb = int(math.ceil(len(ciphertext) / 1024))
        progress = self._create_progress_iterator(range(length_kb), length_kb, "Decrypt        ")
        try:
            plaintext = self._media_cipher.decrypt(ciphertext, ref_key, media_info)
            progress.update(length_kb)
            return plaintext
        except Exception as e:
            progress.set_description("Decrypt Error  ")
            logger.error(e)

        return None

    def _write(self, data, filename):
        length_kb = int(math.ceil(len(data) / 1024))
        progress = self._create_progress_iterator(range(length_kb), length_kb, "Write          ")
        try:
            with open(filename, 'wb') as f:
                f.write(data)
            progress.update(length_kb)
            return filename
        except Exception as e:
            progress.set_description("Write error    ")
            logger.error(e)

        return None

    def _create_unique_filepath(self, filepath):
        file_dir = os.path.dirname(filepath)
        filename = os.path.basename(filepath)
        result_filename = filename
        dissected = os.path.splitext(filename)
        count = 0
        while os.path.exists(os.path.join(file_dir, result_filename)):
            count += 1
            result_filename = "%s_%d%s" % (dissected[0], count, dissected[1])

        return os.path.join(file_dir, result_filename)

    def run(self):
        logger.debug(
            "SinkWorker started, storage_dir=%s" % self._storage_dir
        )
        while True:
            media_message_protocolentity = self._jobs.get()
            if media_message_protocolentity is None:
                sys.exit(0)
            if isinstance(media_message_protocolentity, DownloadableMediaMessageProtocolEntity):
                logger.info(
                    "Processing [url=%s, media_key=%s]" %
                    (media_message_protocolentity.url, base64.b64encode(media_message_protocolentity.media_key))
                )
            else:
                logger.info("Processing %s" % media_message_protocolentity.media_type)

            filedata = None
            fileext = None
            if isinstance(media_message_protocolentity, ImageDownloadableMediaMessageProtocolEntity):
                media_info = MediaCipher.INFO_IMAGE
                filename = "image"
            elif isinstance(media_message_protocolentity, AudioDownloadableMediaMessageProtocolEntity):
                media_info = MediaCipher.INFO_AUDIO
                filename = "ptt" if media_message_protocolentity.ptt else "audio"
            elif isinstance(media_message_protocolentity, VideoDownloadableMediaMessageProtocolEntity):
                media_info = MediaCipher.INFO_VIDEO
                filename = "video"
            elif isinstance(media_message_protocolentity, DocumentDownloadableMediaMessageProtocolEntity):
                media_info = MediaCipher.INFO_DOCUM
                filename = media_message_protocolentity.file_name
            elif isinstance(media_message_protocolentity, ContactMediaMessageProtocolEntity):
                filename = media_message_protocolentity.display_name
                filedata = media_message_protocolentity.vcard
                fileext = "vcard"
            else:
                logger.error("Unsupported Media type: %s" % media_message_protocolentity.__class__)
                sys.exit(1)

            if filedata is None:
                enc_data = self._download(media_message_protocolentity.url)
                if enc_data is None:
                    logger.error("Download failed")
                    sys.exit(1)

                filedata = self._decrypt(enc_data, media_message_protocolentity.media_key, media_info)
                if filedata is None:
                    logger.error("Decrypt failed")
                    sys.exit(1)

            if not isinstance(media_message_protocolentity, DocumentDownloadableMediaMessageProtocolEntity):
                if fileext is None:
                    fileext = media_message_protocolentity.mimetype.split('/')[1].split(';')[0]
                filename_full = "%s.%s" % (filename, fileext)
            else:
                filename_full = filename
            filepath = self._create_unique_filepath(os.path.join(self._storage_dir, filename_full))
            if self._write(filedata, filepath):
                logger.info("Wrote %s" % filepath)
            else:
                sys.exit(1)



from .message_media import MediaMessageProtocolEntity
from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_downloadablemedia \
    import DownloadableMediaMessageAttributes


class DownloadableMediaMessageProtocolEntity(MediaMessageProtocolEntity):
    def __init__(self, media_type, downloadable_media_message_attrs, message_attrs):
        # type: (str, DownloadableMediaMessageAttributes, MessageAttributes)
        super(DownloadableMediaMessageProtocolEntity, self).__init__(
            media_type, downloadable_media_message_attrs, message_attrs
        )
        self.url = downloadable_media_message_attrs.url
        self.mimetype = downloadable_media_message_attrs.mimetype
        self.file_sha256 = downloadable_media_message_attrs.file_sha256
        self.file_length = downloadable_media_message_attrs.file_length
        self.media_key = downloadable_media_message_attrs.media_key

    @property
    def url(self):
        return self.proto.url

    @url.setter
    def url(self, value):
        self.proto.url = value

    @property
    def mimetype(self):
        return self.proto.mimetype

    @mimetype.setter
    def mimetype(self, value):
        self.proto.mimetype = value

    @property
    def file_sha256(self):
        return self.proto.file_sha256

    @file_sha256.setter
    def file_sha256(self, value):
        self.proto.file_sha256 = value

    @property
    def file_length(self):
        return self.proto.file_length

    @file_length.setter
    def file_length(self, value):
        self.proto.file_length = value

    @property
    def media_key(self):
        return self.proto.media_key

    @media_key.setter
    def media_key(self, value):
        self.proto.media_key = value



from yowsup.config.base import config
import logging

logger = logging.getLogger(__name__)


class Config(config.Config):
    def __init__(
            self,
            phone=None,
            cc=None,
            password=None,
            pushname=None,
            id=None,
            mcc=None,
            mnc=None,
            sim_mcc=None,
            sim_mnc=None,
            client_static_keypair=None,
            server_static_public=None,
            expid=None,
            fdid=None,
            edge_routing_info=None,
            chat_dns_domain=None
    ):
        super(Config, self).__init__(1)

        self._phone = str(phone) if phone is not None else None  # type: str
        self._cc = cc  # type: int
        self._password = password  # type: str
        self._pushname = pushname  # type: str
        self._id = id
        self._client_static_keypair = client_static_keypair
        self._server_static_public = server_static_public
        self._expid = expid
        self._fdid = fdid
        self._mcc = mcc
        self._mnc = mnc
        self._sim_mcc = sim_mcc
        self._sim_mnc = sim_mnc
        self._edge_routing_info = edge_routing_info
        self._chat_dns_domain = chat_dns_domain

        if self._password is not None:
            logger.warn("Setting a password in Config is deprecated and not used anymore. "
                        "client_static_keypair is used instead")

    def __str__(self):
        from yowsup.config.v1.serialize import ConfigSerialize
        from yowsup.config.transforms.dict_json import DictJsonTransform
        return DictJsonTransform().transform(ConfigSerialize(self.__class__).serialize(self))

    @property
    def phone(self):
        return self._phone

    @phone.setter
    def phone(self, value):
        self._phone = str(value) if value is not None else None

    @property
    def cc(self):
        return self._cc

    @cc.setter
    def cc(self, value):
        self._cc = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value
        if value is not None:
            logger.warn("Setting a password in Config is deprecated and not used anymore. "
                        "client_static_keypair is used instead")

    @property
    def pushname(self):
        return self._pushname

    @pushname.setter
    def pushname(self, value):
        self._pushname = value

    @property
    def mcc(self):
        return self._mcc

    @mcc.setter
    def mcc(self, value):
        self._mcc = value

    @property
    def mnc(self):
        return self._mnc

    @mnc.setter
    def mnc(self, value):
        self._mnc = value

    @property
    def sim_mcc(self):
        return self._sim_mcc

    @sim_mcc.setter
    def sim_mcc(self, value):
        self._sim_mcc = value

    @property
    def sim_mnc(self):
        return self._sim_mnc

    @sim_mnc.setter
    def sim_mnc(self, value):
        self._sim_mnc = value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def client_static_keypair(self):
        return self._client_static_keypair

    @client_static_keypair.setter
    def client_static_keypair(self, value):
        self._client_static_keypair = value

    @property
    def server_static_public(self):
        return self._server_static_public

    @server_static_public.setter
    def server_static_public(self, value):
        self._server_static_public = value

    @property
    def expid(self):
        return self._expid

    @expid.setter
    def expid(self, value):
        self._expid = value

    @property
    def fdid(self):
        return self._fdid

    @fdid.setter
    def fdid(self, value):
        self._fdid = value

    @property
    def edge_routing_info(self):
        return self._edge_routing_info

    @edge_routing_info.setter
    def edge_routing_info(self, value):
        self._edge_routing_info = value

    @property
    def chat_dns_domain(self):
        return self._chat_dns_domain

    @chat_dns_domain.setter
    def chat_dns_domain(self, value):
        self._chat_dns_domain = value



class Config(object):
    def __init__(self, version):
        self._version = version

    def __contains__(self, item):
        return self[item] is not None

    def __getitem__(self, item):
        return getattr(self, "_%s" % item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def keys(self):
        return [var[1:] for var in vars(self)]

    @property
    def version(self):
        return self._version



from yowsup.layers import YowLayer

import logging
import struct


logger = logging.getLogger(__name__)


class YowNoiseSegmentsLayer(YowLayer):

    PROP_ENABLED = "org.openwhatsapp.yowsup.prop.noise.segmented_enabled"

    def __init__(self):
        super(YowNoiseSegmentsLayer, self).__init__()
        self._read_buffer = bytearray()

    def __str__(self):
        return "Noise Segments Layer"

    def send(self, data):
        if len(data) >= 16777216:
            raise ValueError("data too large to write; length=%d" % len(data))

        if self.getProp(self.PROP_ENABLED, False):
            self.toLower(struct.pack('>I', len(data))[1:])

        self.toLower(data)

    def receive(self, data):
        if self.getProp(self.PROP_ENABLED, False):
            self._read_buffer.extend(data)

            while len(self._read_buffer) > 3:
                read_size = struct.unpack('>I', b"\x00" + self._read_buffer[:3])[0] # type: int
                if len(self._read_buffer) >= (3 + read_size):
                    data = self._read_buffer[3:read_size+3]
                    self._read_buffer = self._read_buffer[3 + read_size:]
                    self.toUpper(bytes(data))
                else:
                    break
        else:
            self.toUpper(data)



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class ProtoProtocolEntity(ProtocolEntity):

    def __init__(self, protoData, mediaType = None):
        super(ProtoProtocolEntity, self).__init__("proto")
        self.mediaType = mediaType
        self.protoData = protoData

    def getProtoData(self):
        return self.protoData

    def getMediaType(self):
        return self.mediaType

    def toProtocolTreeNode(self):
        attribs = {}
        if self.mediaType:
            attribs["mediatype"] = self.mediaType
        return ProtocolTreeNode("proto", attribs, data=self.protoData)

    @staticmethod
    def fromProtocolTreeNode(node):
        return ProtoProtocolEntity(node.data, node["mediatype"])


from yowsup.structs import ProtocolTreeNode
import math
import binascii
import sys
import zlib


class ReadDecoder:
    def __init__(self, tokenDictionary):
        self.tokenDictionary = tokenDictionary

    def getProtocolTreeNode(self, data):
        if type(data) is list:
            data = bytearray(data)

        if data[0] & self.tokenDictionary.FLAG_DEFLATE != 0:
            data = bytearray(b'\x00' + zlib.decompress(bytes(data[1:])))
        if data[0] & self.tokenDictionary.FLAG_SEGMENTED != 0:
            raise ValueError("server to client stanza fragmentation not supported")
        return self.nextTreeInternal(data[1:])

    def getToken(self, index, data):
        token = self.tokenDictionary.getToken(index)
        if not token:
            index = self.readInt8(data)
            token = self.tokenDictionary.getToken(index, True)
            if not token:
                raise ValueError("Invalid token %s" % token)

        return token

    def getTokenDouble(self, n, n2):
        pos = n2 + n * 256
        token = self.tokenDictionary.getToken(pos, True)
        if not token:
            raise ValueError("Invalid token %s" % pos)

        return token

    def streamStart(self, data):
        self.streamStarted = True
        tag = data.pop(0)
        size = self.readListSize(tag, data)
        tag = data.pop(0)

        if tag != 1:
            if tag == 236:
                tag = data.pop(0) + 237
            token = self.getToken(tag, data)#self.tokenDictionary.getToken(tag)
            raise Exception("expecting STREAM_START in streamStart, instead got token: %s" % token)
        attribCount = (size - 2 + size % 2) / 2
        self.readAttributes(attribCount, data)

    def readNibble(self, data):
        _byte = self.readInt8(data)
        ignoreLastNibble = bool(_byte & 0x80)
        size = (_byte & 0x7f)
        nrOfNibbles = size * 2 - int(ignoreLastNibble)
        dataArr = self.readArray(size, data)
        string = ''
        for i in range(0, nrOfNibbles):
            _byte = dataArr[int(math.floor(i/2))]
            _shift = 4 * (1 - i % 2)
            dec = (_byte & (15 << _shift)) >> _shift

            if dec in (0,1,2,3,4,5,6,7,8,9):
                string += str(dec)
            elif dec in (10, 11):
                string += chr(dec - 10 + 45)
            else:
                raise Exception("Bad nibble %s" % dec)
        return string

    def readPacked8(self, n, data):
        size = self.readInt8(data)
        remove = 0
        if (size & 0x80) != 0 and n == 251:
            remove = 1
        size = size & 0x7F
        text = bytearray(self.readArray(size, data))
        hexData = binascii.hexlify(str(text) if sys.version_info < (2,7) else text).upper()
        dataSize = len(hexData)
        out = []
        if remove == 0:
            for i in range(0, dataSize):
                char = chr(hexData[i]) if type(hexData[i]) is int else hexData[i] #python2/3 compat
                val = ord(binascii.unhexlify("0%s" % char))
                if i == (dataSize - 1) and val > 11 and n != 251: continue
                out.append(self.unpackByte(n, val))
        else:
            out =  map(ord, list(hexData[0: -remove])) if sys.version_info < (3,0) else list(hexData[0: -remove])

        return out

    def unpackByte(self, n, n2):
        if n == 251:
            return self.unpackHex(n2)
        if n == 255:
            return self.unpackNibble(n2)
        raise ValueError("bad packed type %s" % n)

    def unpackHex(self, n):
        if n in range(0, 10):
            return n + 48
        if n in range(10, 16):
            return 65 + (n - 10)

        raise ValueError("bad hex %s" % n)

    def unpackNibble(self, n):
        if n in range(0, 10):
            return n + 48
        if n in (10, 11):
            return 45 + (n - 10)
        raise ValueError("bad nibble %s" % n)


    def readHeader(self, data, offset = 0):
        ret = 0
        if len(data) >= (3 + offset):
            b0 = data[offset]
            b1 = data[offset + 1]
            b2 = data[offset + 2]
            ret = b0 + (b1 << 16) + (b2 << 8)

        return ret

    def readInt8(self, data):
        return data.pop(0);

    def readInt16(self, data):
        intTop = data.pop(0)
        intBot = data.pop(0)
        value = (intTop << 8) + intBot
        if value is not None:
            return value
        else:
            return ""

    def readInt20(self, data):
         int1 = data.pop(0)
         int2 = data.pop(0)
         int3 = data.pop(0)
         return ((int1 & 0xF) << 16) | (int2 << 8) | int3

    def readInt24(self, data):
        int1 = data.pop(0)
        int2 = data.pop(0)
        int3 = data.pop(0)
        value = (int1 << 16) + (int2 << 8) + (int3 << 0)
        return value

    def readInt31(self, data):
        data.pop(0)
        int1 = data.pop(0)
        int2 = data.pop(0)
        int3 = data.pop(0)
        return (int1 << 24) | (int1 << 16) | int2 << 8 | int3

    def readListSize(self,token, data):
        size = 0
        if token == 0:
            size = 0
        else:
            if token == 248:
                size = self.readInt8(data)
            else:
                if token == 249:
                    size = self.readInt16(data)
                else:
                    raise Exception("invalid list size in readListSize: token " + str(token))
        return size

    def readAttributes(self, attribCount, data):
        attribs = {}
        for i in range(0, int(attribCount)):
            key = self.readString(self.readInt8(data), data)
            value = self.readString(self.readInt8(data), data)
            attribs[key]=value
        return attribs

    def readString(self,token, data):
        if token == -1:
            raise Exception("-1 token in readString")

        if 2 < token < 236:
            return self.getToken(token, data)

        if token == 0:
            return None

        if token in (236, 237, 238, 239):
            return self.getTokenDouble(token - 236, self.readInt8(data))

        if token == 250:
            user = self.readString(data.pop(0), data)
            server = self.readString(data.pop(0), data)
            if user is not None and server is not None:
                return user + "@" + server
            if server is not None:
                return server
            raise Exception("readString couldn't reconstruct jid")

        if token in (251, 255):
            return "".join(map(chr, self.readPacked8(token, data)))

        if token == 252:
            size8 = self.readInt8(data)
            buf8 = self.readArray(size8, data)
            return "".join(map(chr, buf8))

        if token == 253:
            size20 = self.readInt20(data)
            buf20 = self.readArray(size20, data)
            return "".join(map(chr, buf20))

        if token == 254:
            size31 = self.readInt31()
            buf31 = self.readArray(size31, data)
            return "".join(map(chr, buf31))


        raise Exception("readString couldn't match token "+str(token))

    def readArray(self, length, data):
        out = []
        for i in range(0, length):
            out.append(data.pop(0))

        return out

    def nextTreeInternal(self, data):
        size = self.readListSize(self.readInt8(data), data)
        token = self.readInt8(data)
        if token == 1:
            token = self.readInt8(data)

        if token == 2:
            return None

        tag = self.readString(token, data)

        if size == 0 or tag is None:
            raise ValueError("nextTree sees 0 list or null tag")

        attribCount = (size - 2 + size % 2)/2
        attribs = self.readAttributes(attribCount, data)
        if size % 2 ==1:
            return ProtocolTreeNode(tag, attribs)

        read2 = self.readInt8(data)

        nodeData = None
        nodeChildren = None
        if self.isListTag(read2):
            nodeChildren = self.readList(read2, data)
        elif read2 == 252:
            size = self.readInt8(data)
            nodeData = self.readArray(size, data)
        elif read2 == 253:
            size = self.readInt20(data)
            nodeData = self.readArray(size, data)
        elif read2 == 254:
            size = self.readInt31(data)
            nodeData = self.readArray(size, data)
        elif read2 in (255, 251):
            nodeData = self.readPacked8(read2, data)
        else:
            nodeData = self.readString(read2, data)

        if nodeData and type(nodeData) is not str:
            nodeData = "".join(map(chr, nodeData))

        return ProtocolTreeNode(tag, attribs, nodeChildren, nodeData)

    def readList(self,token, data):
        size = self.readListSize(token, data)
        listx = []
        for i in range(0,size):
            listx.append(self.nextTreeInternal(data))

        return listx;

    def isListTag(self, b):
        return b in (248, 0, 249)



from yowsup.common import YowConstants
from yowsup.layers.protocol_notifications.protocolentities import NotificationProtocolEntity
from yowsup.structs import ProtocolTreeNode


class RequestKeysEncryptNotification(NotificationProtocolEntity):
    """
    <notification t="1419824928" id="2451228097" from="s.whatsapp.net" type="encrypt">
        <count value="9">
        </count>
    </notification>
    """
    def __init__(self, count, timestamp, _id = None, notify = None, offline = None):
        super(RequestKeysEncryptNotification, self).__init__(
            "encrypt", _id, YowConstants.WHATSAPP_SERVER, timestamp, notify, offline
        )
        self._count = count

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, value):
        self._count = value

    def toProtocolTreeNode(self):
        node = super(RequestKeysEncryptNotification, self).toProtocolTreeNode()
        count_node = ProtocolTreeNode("count", {"value": str(self.count)})
        node.addChild(count_node)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = NotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = RequestKeysEncryptNotification
        entity.count = node.getChild("count")["value"]
        return entity



from yowsup.config.base.transform import ConfigTransform


class ConfigDictTransform(ConfigTransform):
    def __init__(self, cls):
        self._cls = cls

    def transform(self, config):
        """
        :param config:
        :type config: dict
        :return:
        :rtype: yowsup.config.config.Config
        """
        out = {}
        for prop in vars(config):
            out[prop] = getattr(config, prop)
        return out

    def reverse(self, data):
        """
        :param data:
        :type data: yowsup,config.config.Config
        :return:
        :rtype: dict
        """
        return self._cls(**data)



from yowsup.config.base.transform import ConfigTransform


class FilterTransform(ConfigTransform):

    def __init__(self, transform_filter=None, reverse_filter=None):
        """
        :param transform_filter:
        :type transform_filter: function | None
        :param reverse_filter:
        :type reverse_filter: function | None
        """
        self._transform_filter = transform_filter  # type: function | None
        self._reverse_filter = reverse_filter  # type: function | None

    def transform(self, data):
        if self._transform_filter is not None:
            out = {}
            for key, val in data.items():
                if self._transform_filter(key, val):
                    out[key] = val
            return out
        return data

    def reverse(self, data):
        if self._reverse_filter is not None:
            out = {}
            for key, val in data.items():
                if self._reverse_filter(key, val):
                    out[key] = val
            return out
        return data



from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_downloadablemedia \
    import DownloadableMediaMessageAttributes
from yowsup.layers.protocol_media.protocolentities.attributes.attributes_image import ImageAttributes
from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes


class ImageDownloadableMediaMessageProtocolEntity(DownloadableMediaMessageProtocolEntity):
    def __init__(self, image_attrs, downloadablemedia_attrs, message_attrs):
        # type: (ImageAttributes, DownloadableMediaMessageAttributes, MessageAttributes) -> None
        super(ImageDownloadableMediaMessageProtocolEntity, self).__init__(
            "image", downloadablemedia_attrs, message_attrs
        )
        self.width = image_attrs.width
        self.height = image_attrs.height
        self.caption = image_attrs.caption
        self.jpeg_thumbnail = image_attrs.jpeg_thumbnail

    @property
    def proto(self):
        return self._proto.image_message

    @property
    def width(self):
        return self.proto.width

    @width.setter
    def width(self, value):
        self.proto.width = value

    @property
    def height(self):
        return self.proto.height

    @height.setter
    def height(self, value):
        self.proto.height = value

    @property
    def jpeg_thumbnail(self):
        return self.proto.jpeg_thumbnail

    @jpeg_thumbnail.setter
    def jpeg_thumbnail(self, value):
        self.proto.jpeg_thumbnail = value if value is not None else b""

    @property
    def caption(self):
        return self.proto.caption

    @caption.setter
    def caption(self, value):
        self.proto.caption = value if value is not None else ""



from yowsup.layers.protocol_media.protocolentities.attributes.attributes_media import MediaAttributes
from yowsup.common.tools import MimeTools
import base64
import hashlib
import os


class DownloadableMediaMessageAttributes(MediaAttributes):
    def __init__(self, mimetype, file_length, file_sha256, url=None, media_key=None, context_info=None):
        super(DownloadableMediaMessageAttributes, self).__init__(context_info)
        self._mimetype = mimetype
        self._file_length = file_length
        self._file_sha256 = file_sha256
        self._url = url
        self._media_key = media_key

    def __str__(self):
        return "[url=%s, mimetype=%s, file_length=%d, file_sha256=%s, media_key=%s]" % (
            self.url,self.mimetype, self.file_length, base64.b64encode(self.file_sha256),
            base64.b64encode(self.media_key)
        )

    @property
    def url(self):
        return self._url

    @property
    def mimetype(self):
        return self._mimetype

    @property
    def file_length(self):
        return self._file_length

    @property
    def file_sha256(self):
        return self._file_sha256

    @property
    def media_key(self):
        return self._media_key

    @staticmethod
    def from_file(
            filepath, mimetype=None, file_length=None,
            file_sha256=None, url=None, media_key=None, context_info=None
    ):
        mimetype = MimeTools.getMIME(filepath) if mimetype is None else mimetype
        file_length = os.path.getsize(filepath) if file_length is None else file_length
        if file_sha256 is None:
            with open(filepath, 'rb') as f:
                file_sha256 = hashlib.sha256(f.read()).digest()
        return DownloadableMediaMessageAttributes(
            mimetype, file_length, file_sha256, url, media_key, context_info
        )



from axolotl.kdf.hkdfv3 import HKDFv3
from axolotl.util.byteutil import ByteUtil

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
import hmac
import hashlib


class MediaCipher(object):
    INFO_IMAGE = b"WhatsApp Image Keys"
    INFO_AUDIO = b"WhatsApp Audio Keys"
    INFO_VIDEO = b"WhatsApp Video Keys"
    INFO_DOCUM = b"WhatsApp Document Keys"

    def encrypt_image(self, plaintext, ref_key):
        return self.encrypt(plaintext, ref_key, self.INFO_IMAGE)

    def encrypt_audio(self, ciphertext, ref_key):
        return self.encrypt(ciphertext, ref_key, self.INFO_AUDIO)

    def encrypt_video(self, ciphertext, ref_key):
        return self.encrypt(ciphertext, ref_key, self.INFO_VIDEO)

    def encrypt_document(self, ciphertext, ref_key):
        return self.encrypt(ciphertext, ref_key, self.INFO_DOCUM)

    def decrypt_image(self, ciphertext, ref_key):
        return self.decrypt(ciphertext, ref_key, self.INFO_IMAGE)

    def decrypt_audio(self, ciphertext, ref_key):
        return self.decrypt(ciphertext, ref_key, self.INFO_AUDIO)

    def decrypt_video(self, ciphertext, ref_key):
        return self.decrypt(ciphertext, ref_key, self.INFO_VIDEO)

    def decrypt_document(self, ciphertext, ref_key):
        return self.decrypt(ciphertext, ref_key, self.INFO_DOCUM)

    def encrypt(self, plaintext, ref_key, media_info):
        derived = HKDFv3().deriveSecrets(ref_key, media_info, 112)
        parts = ByteUtil.split(derived, 16, 32)
        iv = parts[0]
        key = parts[1]
        mac_key = derived[48:80]

        cipher_encryptor = Cipher(
            algorithms.AES(key), modes.CBC(iv), backend=default_backend()
        ).encryptor()
        if len(plaintext) % 16 != 0:
            padder = padding.PKCS7(128).padder()
            padded_plaintext = padder.update(plaintext) + padder.finalize()
        else:
            padded_plaintext = plaintext
        ciphertext = cipher_encryptor.update(padded_plaintext) + cipher_encryptor.finalize()

        mac = hmac.new(mac_key, digestmod=hashlib.sha256)
        mac.update(iv)
        mac.update(ciphertext)

        return ciphertext + mac.digest()[:10]

    def decrypt(self, ciphertext, ref_key, media_info):
        derived = HKDFv3().deriveSecrets(ref_key, media_info, 112)
        parts = ByteUtil.split(derived, 16, 32)
        iv = parts[0]
        key = parts[1]
        mac_key = derived[48:80]
        media_ciphertext = ciphertext[:-10]
        mac_value = ciphertext[-10:]

        mac = hmac.new(mac_key, digestmod=hashlib.sha256)
        mac.update(iv)
        mac.update(media_ciphertext)

        if mac_value != mac.digest()[:10]:
            raise ValueError("Invalid MAC")

        cipher_decryptor = Cipher(
            algorithms.AES(key), modes.CBC(iv), backend=default_backend()
        ).decryptor()

        decrypted = cipher_decryptor.update(media_ciphertext) + cipher_decryptor.finalize()
        unpadder = padding.PKCS7(128).unpadder()
        return unpadder.update(decrypted) + unpadder.finalize()



from yowsup.layers.protocol_messages.protocolentities.attributes.attributes_message import MessageAttributes
from .message_media import MediaMessageProtocolEntity
from .attributes.attributes_extendedtext import ExtendedTextAttributes
from .attributes.attributes_media import MediaAttributes


class ExtendedTextMediaMessageProtocolEntity(MediaMessageProtocolEntity):
    def __init__(self, extended_text_attrs, media_attrs, message_attrs):
        # type: (ExtendedTextAttributes, MediaAttributes, MessageAttributes) -> None
        super(ExtendedTextMediaMessageProtocolEntity, self).__init__("url", media_attrs, message_attrs)
        self.text = extended_text_attrs.text
        self.matched_text = extended_text_attrs.matched_text
        self.canonical_url = extended_text_attrs.canonical_url
        self.description = extended_text_attrs.description
        self.title = extended_text_attrs.title
        self.jpeg_thumbnail = extended_text_attrs.jpeg_thumbnail

    @property
    def proto(self):
        return self._proto.extended_text_message

    @property
    def text(self):
        return self.proto.text

    @text.setter
    def text(self, value):
        self.proto.text = value

    @property
    def matched_text(self):
        return self.proto.matched_text

    @matched_text.setter
    def matched_text(self, value):
        self.proto.matched_text = value

    @property
    def canonical_url(self):
        return self.proto.canonical_url

    @canonical_url.setter
    def canonical_url(self, value):
        self.proto.canonical_url = value

    @property
    def description(self):
        return self.proto.description

    @description.setter
    def description(self, value):
        self.proto.description = value

    @property
    def title(self):
        return self.proto.title

    @title.setter
    def title(self, value):
        self.proto.title = value

    @property
    def jpeg_thumbnail(self):
        return self.proto.jpeg_thumbnail

    @jpeg_thumbnail.setter
    def jpeg_thumbnail(self, value):
        self.proto.jpeg_thumbnail = value



class ExtendedTextAttributes(object):
    def __init__(self,
                 text, matched_text, canonical_url, description, title, jpeg_thumbnail
                 ):
        self._text = text
        self._matched_text = matched_text
        self._canonical_url = canonical_url
        self._description = description
        self._title = title
        self._jpeg_thumbnail = jpeg_thumbnail

    @property
    def text(self):
        return self._text

    @property
    def matched_text(self):
        return self._matched_text

    @property
    def canonical_url(self):
        return self._canonical_url

    @property
    def description(self):
        return self._description

    @property
    def title(self):
        return self._title

    @property
    def jpeg_thumbnail(self):
        return self._jpeg_thumbnail



from yowsup.config.transforms.props import PropsTransform


class SerializeTransform(PropsTransform):

    def __init__(self, serialize_map):
        """
        {
            "keystore": serializer
        }
        :param serialize_map:
        :type serialize_map:
        """
        transform_map = {}
        reverse_map = {}
        for key, val in serialize_map:
            transform_map[key] = lambda key, val: key, serialize_map[key].serialize(val)
            reverse_map[key] = lambda key, val: key, serialize_map[key].deserialize(val)

        super(SerializeTransform, self).__init__(transform_map=transform_map, reverse_map=reverse_map)




from yowsup.layers.interface import YowInterfaceLayer, ProtocolEntityCallback
from yowsup.layers.protocol_media.protocolentities import *
from yowsup.layers.network.layer import YowNetworkLayer
from yowsup.layers import EventCallback
import sys

try:
    from tqdm import tqdm
    import requests
except ImportError:
    print("This demo requires the python packages 'tqdm' and 'requests' to be installed, exiting.")
    sys.exit(1)

from ..common.sink_worker import SinkWorker

import tempfile
import logging
import os

logger = logging.getLogger(__name__)


class MediaSinkLayer(YowInterfaceLayer):
    PROP_STORAGE_DIR = "org.openwhatsapp.yowsup.prop.demos.mediasink.storage_dir"

    def __init__(self):
        super(MediaSinkLayer, self).__init__()
        self._sink_worker = None

    @EventCallback(YowNetworkLayer.EVENT_STATE_CONNECTED)
    def on_connected(self, event):
        logger.info("Connected, starting SinkWorker")
        storage_dir = self.getProp(self.PROP_STORAGE_DIR)
        if storage_dir is None:
            logger.debug("No storage dir specified, creating tempdir")
            storage_dir = tempfile.mkdtemp("yowsup_mediasink")

        if not os.path.exists(storage_dir):
            logger.debug("%s does not exist, creating" % storage_dir)
            os.makedirs(storage_dir)

        logger.info("Storing incoming media to %s" % storage_dir)
        self._sink_worker = SinkWorker(storage_dir)
        self._sink_worker.start()

    @ProtocolEntityCallback("message")
    def on_message(self, message_protocolentity):
        self.toLower(message_protocolentity.ack())
        self.toLower(message_protocolentity.ack(True))
        if isinstance(message_protocolentity, MediaMessageProtocolEntity):
            self.on_media_message(message_protocolentity)

    @ProtocolEntityCallback("receipt")
    def on_receipt(self, entity):
        self.toLower(entity.ack())

    def on_media_message(self, media_message_protocolentity):
        self._sink_worker.enqueue(media_message_protocolentity)




from .message_media import MediaMessageProtocolEntity
from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
from .message_media_downloadable_image import ImageDownloadableMediaMessageProtocolEntity
from .message_media_downloadable_audio import AudioDownloadableMediaMessageProtocolEntity
from .message_media_downloadable_video import VideoDownloadableMediaMessageProtocolEntity
from .message_media_downloadable_document import DocumentDownloadableMediaMessageProtocolEntity
from .message_media_location import LocationMediaMessageProtocolEntity
from .message_media_contact import ContactMediaMessageProtocolEntity
from .message_media_extendedtext import ExtendedTextMediaMessageProtocolEntity
from .iq_requestupload import RequestUploadIqProtocolEntity
from .iq_requestupload_result import ResultRequestUploadIqProtocolEntity



from yowsup.common.http.warequest import WARequest
from yowsup.common.http.waresponseparser import JSONResponseParser
from yowsup.common.tools import WATools
from yowsup.registration.existsrequest import WAExistsRequest
from yowsup.env import YowsupEnv


class WACodeRequest(WARequest):
    def __init__(self, method, config):
        """
        :type method: str
        :param config:
        :type config: yowsup.config.v1.config.Config
        """
        super(WACodeRequest,self).__init__(config)

        self.addParam("mcc", config.mcc.zfill(3))
        self.addParam("mnc", config.mnc.zfill(3))
        self.addParam("sim_mcc", config.sim_mcc.zfill(3))
        self.addParam("sim_mnc", config.sim_mnc.zfill(3))
        self.addParam("method", method)
        self.addParam("reason", "")
        self.addParam("token", YowsupEnv.getCurrent().getToken(self._p_in))
        self.addParam("hasav", "1")

        self.url = "v.whatsapp.net/v2/code"

        self.pvars = ["status","reason","length", "method", "retry_after", "code", "param"] +\
                    ["login", "type", "sms_wait", "voice_wait"]
        self.setParser(JSONResponseParser())

    def send(self, parser = None, encrypt=True, preview=False):
        if self. _config.id is not None:
            request = WAExistsRequest(self._config)
            result = request.send(encrypt=encrypt, preview=preview)

            if result:
                if result["status"] == "ok":
                    return result
                elif result["status"] == "fail" and "reason" in result and result["reason"] == "blocked":
                    return result
        else:
            self._config.id = WATools.generateIdentity()
            self.addParam("id", self._config.id)

        res = super(WACodeRequest, self).send(parser, encrypt=encrypt, preview=preview)

        return res



class MessageAttributes(object):
    def __init__(
            self, id=None, sender=None, recipient=None, notify=None, timestamp=None, participant=None, offline=None,
            retry=None
    ):
        assert (sender or recipient), "Must specify either sender or recipient " \
                                      "jid to create the message "
        assert not (sender and recipient), "Can't set both attributes to message at same " \
                                           "time (sender, recipient) "
        self.id = id
        self.sender = sender
        self.recipient = recipient
        self.notify = notify
        self.timestamp = int(timestamp) if timestamp else None
        self.participant = participant
        self.offline = offline in ("1", True)
        self.retry = int(retry) if retry else None

    @staticmethod
    def from_message_protocoltreenode(node):
        return MessageAttributes(
            node["id"], node["from"], node["to"], node["notify"], node["t"], node["participant"], node["offline"],
            node["retry"]
        )



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: e2e.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import protocol_pb2 as protocol__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='e2e.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\te2e.proto\x1a\x0eprotocol.proto\"\xb0\x01\n\x0b\x43ontextInfo\x12\x11\n\tstanza_id\x18\x01 \x01(\t\x12\x13\n\x0bparticipant\x18\x02 \x01(\t\x12 \n\x0equoted_message\x18\x03 \x01(\x0b\x32\x08.Message\x12\x12\n\nremote_jid\x18\x04 \x01(\t\x12\x15\n\rmentioned_jid\x18\x0f \x03(\t\x12\x14\n\x0c\x65\x64it_version\x18\x10 \x01(\r\x12\x16\n\x0erevoke_message\x18\x11 \x01(\x08\"\xce\x16\n\x07Message\x12\x14\n\x0c\x63onversation\x18\x01 \x01(\t\x12N\n\x1fsender_key_distribution_message\x18\x02 \x01(\x0b\x32%.Message.SenderKeyDistributionMessage\x12,\n\rimage_message\x18\x03 \x01(\x0b\x32\x15.Message.ImageMessage\x12\x30\n\x0f\x63ontact_message\x18\x04 \x01(\x0b\x32\x17.Message.ContactMessage\x12\x32\n\x10location_message\x18\x05 \x01(\x0b\x32\x18.Message.LocationMessage\x12;\n\x15\x65xtended_text_message\x18\x06 \x01(\x0b\x32\x1c.Message.ExtendedTextMessage\x12\x32\n\x10\x64ocument_message\x18\x07 \x01(\x0b\x32\x18.Message.DocumentMessage\x12,\n\raudio_message\x18\x08 \x01(\x0b\x32\x15.Message.AudioMessage\x12,\n\rvideo_message\x18\t \x01(\x0b\x32\x15.Message.VideoMessage\x12\x1b\n\x04\x63\x61ll\x18\n \x01(\x0b\x32\r.Message.Call\x12\x1b\n\x04\x63hat\x18\x0b \x01(\x0b\x32\r.Message.Chat\x12\x32\n\x10protocol_message\x18\x0c \x01(\x0b\x32\x18.Message.ProtocolMessage\x12=\n\x16\x63ontacts_array_message\x18\r \x01(\x0b\x32\x1d.Message.ContactsArrayMessage\x12\x43\n\x19highly_structured_message\x18\x0e \x01(\x0b\x32 .Message.HighlyStructuredMessage\x1a\x61\n\x1cSenderKeyDistributionMessage\x12\x10\n\x08group_id\x18\x01 \x01(\t\x12/\n\'axolotl_sender_key_distribution_message\x18\x02 \x01(\x0c\x1a\xd6\x01\n\x0cImageMessage\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x10\n\x08mimetype\x18\x02 \x01(\t\x12\x0f\n\x07\x63\x61ption\x18\x03 \x01(\t\x12\x13\n\x0b\x66ile_sha256\x18\x04 \x01(\x0c\x12\x13\n\x0b\x66ile_length\x18\x05 \x01(\x04\x12\x0e\n\x06height\x18\x06 \x01(\r\x12\r\n\x05width\x18\x07 \x01(\r\x12\x11\n\tmedia_key\x18\x08 \x01(\x0c\x12\x16\n\x0ejpeg_thumbnail\x18\x10 \x01(\x0c\x12\"\n\x0c\x63ontext_info\x18\x11 \x01(\x0b\x32\x0c.ContextInfo\x1aY\n\x0e\x43ontactMessage\x12\x14\n\x0c\x64isplay_name\x18\x01 \x01(\t\x12\r\n\x05vcard\x18\x10 \x01(\x0c\x12\"\n\x0c\x63ontext_info\x18\x11 \x01(\x0b\x32\x0c.ContextInfo\x1a\xd2\x02\n\x0fLocationMessage\x12\x18\n\x10\x64\x65grees_latitude\x18\x01 \x01(\x01\x12\x19\n\x11\x64\x65grees_longitude\x18\x02 \x01(\x01\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x04 \x01(\t\x12\x0b\n\x03url\x18\x05 \x01(\t\x12\x10\n\x08\x64uration\x18\x06 \x01(\x02\x12\x1a\n\x12\x61\x63\x63uracy_in_meters\x18\x07 \x01(\r\x12\x14\n\x0cspeed_in_mps\x18\x08 \x01(\x02\x12-\n%degrees_clockwise_from_magnetic_north\x18\t \x01(\r\x12/\n\'axolotl_sender_key_distribution_message\x18\n \x01(\x0c\x12\x16\n\x0ejpeg_thumbnail\x18\x10 \x01(\x0c\x12\"\n\x0c\x63ontext_info\x18\x11 \x01(\x0b\x32\x0c.ContextInfo\x1a\xb0\x01\n\x13\x45xtendedTextMessage\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x14\n\x0cmatched_text\x18\x02 \x01(\t\x12\x15\n\rcanonical_url\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\r\n\x05title\x18\x06 \x01(\t\x12\x16\n\x0ejpeg_thumbnail\x18\x10 \x01(\x0c\x12\"\n\x0c\x63ontext_info\x18\x11 \x01(\x0b\x32\x0c.ContextInfo\x1a\xdf\x01\n\x0f\x44ocumentMessage\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x10\n\x08mimetype\x18\x02 \x01(\t\x12\r\n\x05title\x18\x03 \x01(\t\x12\x13\n\x0b\x66ile_sha256\x18\x04 \x01(\x0c\x12\x13\n\x0b\x66ile_length\x18\x05 \x01(\x04\x12\x12\n\npage_count\x18\x06 \x01(\r\x12\x11\n\tmedia_key\x18\x07 \x01(\x0c\x12\x11\n\tfile_name\x18\x08 \x01(\t\x12\x16\n\x0ejpeg_thumbnail\x18\x10 \x01(\x0c\x12\"\n\x0c\x63ontext_info\x18\x11 \x01(\x0b\x32\x0c.ContextInfo\x1a\xc7\x01\n\x0c\x41udioMessage\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x10\n\x08mimetype\x18\x02 \x01(\t\x12\x13\n\x0b\x66ile_sha256\x18\x03 \x01(\x0c\x12\x13\n\x0b\x66ile_length\x18\x04 \x01(\x04\x12\x0f\n\x07seconds\x18\x05 \x01(\r\x12\x0b\n\x03ptt\x18\x06 \x01(\x08\x12\x11\n\tmedia_key\x18\x07 \x01(\x0c\x12\"\n\x0c\x63ontext_info\x18\x11 \x01(\x0b\x32\x0c.ContextInfo\x12\x19\n\x11streaming_sidecar\x18\x12 \x01(\x0c\x1a\x89\x03\n\x0cVideoMessage\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x10\n\x08mimetype\x18\x02 \x01(\t\x12\x13\n\x0b\x66ile_sha256\x18\x03 \x01(\x0c\x12\x13\n\x0b\x66ile_length\x18\x04 \x01(\x04\x12\x0f\n\x07seconds\x18\x05 \x01(\r\x12\x11\n\tmedia_key\x18\x06 \x01(\x0c\x12\x0f\n\x07\x63\x61ption\x18\x07 \x01(\t\x12\x14\n\x0cgif_playback\x18\x08 \x01(\x08\x12\x0e\n\x06height\x18\t \x01(\r\x12\r\n\x05width\x18\n \x01(\r\x12\x16\n\x0ejpeg_thumbnail\x18\x10 \x01(\x0c\x12\"\n\x0c\x63ontext_info\x18\x11 \x01(\x0b\x32\x0c.ContextInfo\x12\x19\n\x11streaming_sidecar\x18\x12 \x01(\x0c\x12@\n\x0fgif_attribution\x18\x13 \x01(\x0e\x32!.Message.VideoMessage.Attribution:\x04NONE\"-\n\x0b\x41ttribution\x12\x08\n\x04NONE\x10\x00\x12\t\n\x05GIPHY\x10\x01\x12\t\n\x05TENOR\x10\x02\x1a\x18\n\x04\x43\x61ll\x12\x10\n\x08\x63\x61ll_key\x18\x01 \x01(\x0c\x1a(\n\x04\x43hat\x12\x14\n\x0c\x64isplay_name\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x1at\n\x0fProtocolMessage\x12\x18\n\x03key\x18\x01 \x01(\x0b\x32\x0b.MessageKey\x12\x33\n\x04type\x18\x02 \x01(\x0e\x32\x1d.Message.ProtocolMessage.Type:\x06REVOKE\"\x12\n\x04Type\x12\n\n\x06REVOKE\x10\x00\x1a{\n\x14\x43ontactsArrayMessage\x12\x14\n\x0c\x64isplay_name\x18\x01 \x01(\t\x12)\n\x08\x63ontacts\x18\x02 \x03(\x0b\x32\x17.Message.ContactMessage\x12\"\n\x0c\x63ontext_info\x18\x11 \x01(\x0b\x32\x0c.ContextInfo\x1a|\n\x17HighlyStructuredMessage\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x14\n\x0c\x65lement_name\x18\x02 \x01(\t\x12\x0e\n\x06params\x18\x03 \x03(\t\x12\x13\n\x0b\x66\x61llback_lg\x18\x04 \x01(\t\x12\x13\n\x0b\x66\x61llback_lc\x18\x05 \x01(\t')
  ,
  dependencies=[protocol__pb2.DESCRIPTOR,])



_MESSAGE_VIDEOMESSAGE_ATTRIBUTION = _descriptor.EnumDescriptor(
  name='Attribution',
  full_name='Message.VideoMessage.Attribution',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GIPHY', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENOR', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2621,
  serialized_end=2666,
)
_sym_db.RegisterEnumDescriptor(_MESSAGE_VIDEOMESSAGE_ATTRIBUTION)

_MESSAGE_PROTOCOLMESSAGE_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='Message.ProtocolMessage.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REVOKE', index=0, number=0,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2834,
  serialized_end=2852,
)
_sym_db.RegisterEnumDescriptor(_MESSAGE_PROTOCOLMESSAGE_TYPE)


_CONTEXTINFO = _descriptor.Descriptor(
  name='ContextInfo',
  full_name='ContextInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='stanza_id', full_name='ContextInfo.stanza_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='participant', full_name='ContextInfo.participant', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quoted_message', full_name='ContextInfo.quoted_message', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='remote_jid', full_name='ContextInfo.remote_jid', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mentioned_jid', full_name='ContextInfo.mentioned_jid', index=4,
      number=15, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='edit_version', full_name='ContextInfo.edit_version', index=5,
      number=16, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='revoke_message', full_name='ContextInfo.revoke_message', index=6,
      number=17, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30,
  serialized_end=206,
)


_MESSAGE_SENDERKEYDISTRIBUTIONMESSAGE = _descriptor.Descriptor(
  name='SenderKeyDistributionMessage',
  full_name='Message.SenderKeyDistributionMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='group_id', full_name='Message.SenderKeyDistributionMessage.group_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='axolotl_sender_key_distribution_message', full_name='Message.SenderKeyDistributionMessage.axolotl_sender_key_distribution_message', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=917,
  serialized_end=1014,
)

_MESSAGE_IMAGEMESSAGE = _descriptor.Descriptor(
  name='ImageMessage',
  full_name='Message.ImageMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='Message.ImageMessage.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mimetype', full_name='Message.ImageMessage.mimetype', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='caption', full_name='Message.ImageMessage.caption', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_sha256', full_name='Message.ImageMessage.file_sha256', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_length', full_name='Message.ImageMessage.file_length', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='Message.ImageMessage.height', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='Message.ImageMessage.width', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='media_key', full_name='Message.ImageMessage.media_key', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='jpeg_thumbnail', full_name='Message.ImageMessage.jpeg_thumbnail', index=8,
      number=16, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_info', full_name='Message.ImageMessage.context_info', index=9,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1017,
  serialized_end=1231,
)

_MESSAGE_CONTACTMESSAGE = _descriptor.Descriptor(
  name='ContactMessage',
  full_name='Message.ContactMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_name', full_name='Message.ContactMessage.display_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vcard', full_name='Message.ContactMessage.vcard', index=1,
      number=16, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_info', full_name='Message.ContactMessage.context_info', index=2,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1233,
  serialized_end=1322,
)

_MESSAGE_LOCATIONMESSAGE = _descriptor.Descriptor(
  name='LocationMessage',
  full_name='Message.LocationMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='degrees_latitude', full_name='Message.LocationMessage.degrees_latitude', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='degrees_longitude', full_name='Message.LocationMessage.degrees_longitude', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='Message.LocationMessage.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='address', full_name='Message.LocationMessage.address', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='url', full_name='Message.LocationMessage.url', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='duration', full_name='Message.LocationMessage.duration', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='accuracy_in_meters', full_name='Message.LocationMessage.accuracy_in_meters', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='speed_in_mps', full_name='Message.LocationMessage.speed_in_mps', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='degrees_clockwise_from_magnetic_north', full_name='Message.LocationMessage.degrees_clockwise_from_magnetic_north', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='axolotl_sender_key_distribution_message', full_name='Message.LocationMessage.axolotl_sender_key_distribution_message', index=9,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='jpeg_thumbnail', full_name='Message.LocationMessage.jpeg_thumbnail', index=10,
      number=16, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_info', full_name='Message.LocationMessage.context_info', index=11,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1325,
  serialized_end=1663,
)

_MESSAGE_EXTENDEDTEXTMESSAGE = _descriptor.Descriptor(
  name='ExtendedTextMessage',
  full_name='Message.ExtendedTextMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='Message.ExtendedTextMessage.text', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='matched_text', full_name='Message.ExtendedTextMessage.matched_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='canonical_url', full_name='Message.ExtendedTextMessage.canonical_url', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='Message.ExtendedTextMessage.description', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='title', full_name='Message.ExtendedTextMessage.title', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='jpeg_thumbnail', full_name='Message.ExtendedTextMessage.jpeg_thumbnail', index=5,
      number=16, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_info', full_name='Message.ExtendedTextMessage.context_info', index=6,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1666,
  serialized_end=1842,
)

_MESSAGE_DOCUMENTMESSAGE = _descriptor.Descriptor(
  name='DocumentMessage',
  full_name='Message.DocumentMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='Message.DocumentMessage.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mimetype', full_name='Message.DocumentMessage.mimetype', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='title', full_name='Message.DocumentMessage.title', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_sha256', full_name='Message.DocumentMessage.file_sha256', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_length', full_name='Message.DocumentMessage.file_length', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_count', full_name='Message.DocumentMessage.page_count', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='media_key', full_name='Message.DocumentMessage.media_key', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='Message.DocumentMessage.file_name', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='jpeg_thumbnail', full_name='Message.DocumentMessage.jpeg_thumbnail', index=8,
      number=16, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_info', full_name='Message.DocumentMessage.context_info', index=9,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1845,
  serialized_end=2068,
)

_MESSAGE_AUDIOMESSAGE = _descriptor.Descriptor(
  name='AudioMessage',
  full_name='Message.AudioMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='Message.AudioMessage.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mimetype', full_name='Message.AudioMessage.mimetype', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_sha256', full_name='Message.AudioMessage.file_sha256', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_length', full_name='Message.AudioMessage.file_length', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='seconds', full_name='Message.AudioMessage.seconds', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ptt', full_name='Message.AudioMessage.ptt', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='media_key', full_name='Message.AudioMessage.media_key', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_info', full_name='Message.AudioMessage.context_info', index=7,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='streaming_sidecar', full_name='Message.AudioMessage.streaming_sidecar', index=8,
      number=18, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2071,
  serialized_end=2270,
)

_MESSAGE_VIDEOMESSAGE = _descriptor.Descriptor(
  name='VideoMessage',
  full_name='Message.VideoMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='Message.VideoMessage.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mimetype', full_name='Message.VideoMessage.mimetype', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_sha256', full_name='Message.VideoMessage.file_sha256', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_length', full_name='Message.VideoMessage.file_length', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='seconds', full_name='Message.VideoMessage.seconds', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='media_key', full_name='Message.VideoMessage.media_key', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='caption', full_name='Message.VideoMessage.caption', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gif_playback', full_name='Message.VideoMessage.gif_playback', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='Message.VideoMessage.height', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='Message.VideoMessage.width', index=9,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='jpeg_thumbnail', full_name='Message.VideoMessage.jpeg_thumbnail', index=10,
      number=16, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_info', full_name='Message.VideoMessage.context_info', index=11,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='streaming_sidecar', full_name='Message.VideoMessage.streaming_sidecar', index=12,
      number=18, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gif_attribution', full_name='Message.VideoMessage.gif_attribution', index=13,
      number=19, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MESSAGE_VIDEOMESSAGE_ATTRIBUTION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2273,
  serialized_end=2666,
)

_MESSAGE_CALL = _descriptor.Descriptor(
  name='Call',
  full_name='Message.Call',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='call_key', full_name='Message.Call.call_key', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2668,
  serialized_end=2692,
)

_MESSAGE_CHAT = _descriptor.Descriptor(
  name='Chat',
  full_name='Message.Chat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_name', full_name='Message.Chat.display_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='Message.Chat.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2694,
  serialized_end=2734,
)

_MESSAGE_PROTOCOLMESSAGE = _descriptor.Descriptor(
  name='ProtocolMessage',
  full_name='Message.ProtocolMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Message.ProtocolMessage.key', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='Message.ProtocolMessage.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MESSAGE_PROTOCOLMESSAGE_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2736,
  serialized_end=2852,
)

_MESSAGE_CONTACTSARRAYMESSAGE = _descriptor.Descriptor(
  name='ContactsArrayMessage',
  full_name='Message.ContactsArrayMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_name', full_name='Message.ContactsArrayMessage.display_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contacts', full_name='Message.ContactsArrayMessage.contacts', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_info', full_name='Message.ContactsArrayMessage.context_info', index=2,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2854,
  serialized_end=2977,
)

_MESSAGE_HIGHLYSTRUCTUREDMESSAGE = _descriptor.Descriptor(
  name='HighlyStructuredMessage',
  full_name='Message.HighlyStructuredMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='namespace', full_name='Message.HighlyStructuredMessage.namespace', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='element_name', full_name='Message.HighlyStructuredMessage.element_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='params', full_name='Message.HighlyStructuredMessage.params', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fallback_lg', full_name='Message.HighlyStructuredMessage.fallback_lg', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fallback_lc', full_name='Message.HighlyStructuredMessage.fallback_lc', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2979,
  serialized_end=3103,
)

_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='conversation', full_name='Message.conversation', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sender_key_distribution_message', full_name='Message.sender_key_distribution_message', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='image_message', full_name='Message.image_message', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contact_message', full_name='Message.contact_message', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='location_message', full_name='Message.location_message', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extended_text_message', full_name='Message.extended_text_message', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='document_message', full_name='Message.document_message', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='audio_message', full_name='Message.audio_message', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='video_message', full_name='Message.video_message', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='call', full_name='Message.call', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chat', full_name='Message.chat', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='protocol_message', full_name='Message.protocol_message', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contacts_array_message', full_name='Message.contacts_array_message', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='highly_structured_message', full_name='Message.highly_structured_message', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_MESSAGE_SENDERKEYDISTRIBUTIONMESSAGE, _MESSAGE_IMAGEMESSAGE, _MESSAGE_CONTACTMESSAGE, _MESSAGE_LOCATIONMESSAGE, _MESSAGE_EXTENDEDTEXTMESSAGE, _MESSAGE_DOCUMENTMESSAGE, _MESSAGE_AUDIOMESSAGE, _MESSAGE_VIDEOMESSAGE, _MESSAGE_CALL, _MESSAGE_CHAT, _MESSAGE_PROTOCOLMESSAGE, _MESSAGE_CONTACTSARRAYMESSAGE, _MESSAGE_HIGHLYSTRUCTUREDMESSAGE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=209,
  serialized_end=3103,
)

_CONTEXTINFO.fields_by_name['quoted_message'].message_type = _MESSAGE
_MESSAGE_SENDERKEYDISTRIBUTIONMESSAGE.containing_type = _MESSAGE
_MESSAGE_IMAGEMESSAGE.fields_by_name['context_info'].message_type = _CONTEXTINFO
_MESSAGE_IMAGEMESSAGE.containing_type = _MESSAGE
_MESSAGE_CONTACTMESSAGE.fields_by_name['context_info'].message_type = _CONTEXTINFO
_MESSAGE_CONTACTMESSAGE.containing_type = _MESSAGE
_MESSAGE_LOCATIONMESSAGE.fields_by_name['context_info'].message_type = _CONTEXTINFO
_MESSAGE_LOCATIONMESSAGE.containing_type = _MESSAGE
_MESSAGE_EXTENDEDTEXTMESSAGE.fields_by_name['context_info'].message_type = _CONTEXTINFO
_MESSAGE_EXTENDEDTEXTMESSAGE.containing_type = _MESSAGE
_MESSAGE_DOCUMENTMESSAGE.fields_by_name['context_info'].message_type = _CONTEXTINFO
_MESSAGE_DOCUMENTMESSAGE.containing_type = _MESSAGE
_MESSAGE_AUDIOMESSAGE.fields_by_name['context_info'].message_type = _CONTEXTINFO
_MESSAGE_AUDIOMESSAGE.containing_type = _MESSAGE
_MESSAGE_VIDEOMESSAGE.fields_by_name['context_info'].message_type = _CONTEXTINFO
_MESSAGE_VIDEOMESSAGE.fields_by_name['gif_attribution'].enum_type = _MESSAGE_VIDEOMESSAGE_ATTRIBUTION
_MESSAGE_VIDEOMESSAGE.containing_type = _MESSAGE
_MESSAGE_VIDEOMESSAGE_ATTRIBUTION.containing_type = _MESSAGE_VIDEOMESSAGE
_MESSAGE_CALL.containing_type = _MESSAGE
_MESSAGE_CHAT.containing_type = _MESSAGE
_MESSAGE_PROTOCOLMESSAGE.fields_by_name['key'].message_type = protocol__pb2._MESSAGEKEY
_MESSAGE_PROTOCOLMESSAGE.fields_by_name['type'].enum_type = _MESSAGE_PROTOCOLMESSAGE_TYPE
_MESSAGE_PROTOCOLMESSAGE.containing_type = _MESSAGE
_MESSAGE_PROTOCOLMESSAGE_TYPE.containing_type = _MESSAGE_PROTOCOLMESSAGE
_MESSAGE_CONTACTSARRAYMESSAGE.fields_by_name['contacts'].message_type = _MESSAGE_CONTACTMESSAGE
_MESSAGE_CONTACTSARRAYMESSAGE.fields_by_name['context_info'].message_type = _CONTEXTINFO
_MESSAGE_CONTACTSARRAYMESSAGE.containing_type = _MESSAGE
_MESSAGE_HIGHLYSTRUCTUREDMESSAGE.containing_type = _MESSAGE
_MESSAGE.fields_by_name['sender_key_distribution_message'].message_type = _MESSAGE_SENDERKEYDISTRIBUTIONMESSAGE
_MESSAGE.fields_by_name['image_message'].message_type = _MESSAGE_IMAGEMESSAGE
_MESSAGE.fields_by_name['contact_message'].message_type = _MESSAGE_CONTACTMESSAGE
_MESSAGE.fields_by_name['location_message'].message_type = _MESSAGE_LOCATIONMESSAGE
_MESSAGE.fields_by_name['extended_text_message'].message_type = _MESSAGE_EXTENDEDTEXTMESSAGE
_MESSAGE.fields_by_name['document_message'].message_type = _MESSAGE_DOCUMENTMESSAGE
_MESSAGE.fields_by_name['audio_message'].message_type = _MESSAGE_AUDIOMESSAGE
_MESSAGE.fields_by_name['video_message'].message_type = _MESSAGE_VIDEOMESSAGE
_MESSAGE.fields_by_name['call'].message_type = _MESSAGE_CALL
_MESSAGE.fields_by_name['chat'].message_type = _MESSAGE_CHAT
_MESSAGE.fields_by_name['protocol_message'].message_type = _MESSAGE_PROTOCOLMESSAGE
_MESSAGE.fields_by_name['contacts_array_message'].message_type = _MESSAGE_CONTACTSARRAYMESSAGE
_MESSAGE.fields_by_name['highly_structured_message'].message_type = _MESSAGE_HIGHLYSTRUCTUREDMESSAGE
DESCRIPTOR.message_types_by_name['ContextInfo'] = _CONTEXTINFO
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ContextInfo = _reflection.GeneratedProtocolMessageType('ContextInfo', (_message.Message,), dict(
  DESCRIPTOR = _CONTEXTINFO,
  __module__ = 'e2e_pb2'
  # @@protoc_insertion_point(class_scope:ContextInfo)
  ))
_sym_db.RegisterMessage(ContextInfo)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), dict(

  SenderKeyDistributionMessage = _reflection.GeneratedProtocolMessageType('SenderKeyDistributionMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_SENDERKEYDISTRIBUTIONMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.SenderKeyDistributionMessage)
    ))
  ,

  ImageMessage = _reflection.GeneratedProtocolMessageType('ImageMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_IMAGEMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.ImageMessage)
    ))
  ,

  ContactMessage = _reflection.GeneratedProtocolMessageType('ContactMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_CONTACTMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.ContactMessage)
    ))
  ,

  LocationMessage = _reflection.GeneratedProtocolMessageType('LocationMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_LOCATIONMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.LocationMessage)
    ))
  ,

  ExtendedTextMessage = _reflection.GeneratedProtocolMessageType('ExtendedTextMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_EXTENDEDTEXTMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.ExtendedTextMessage)
    ))
  ,

  DocumentMessage = _reflection.GeneratedProtocolMessageType('DocumentMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_DOCUMENTMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.DocumentMessage)
    ))
  ,

  AudioMessage = _reflection.GeneratedProtocolMessageType('AudioMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_AUDIOMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.AudioMessage)
    ))
  ,

  VideoMessage = _reflection.GeneratedProtocolMessageType('VideoMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_VIDEOMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.VideoMessage)
    ))
  ,

  Call = _reflection.GeneratedProtocolMessageType('Call', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_CALL,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.Call)
    ))
  ,

  Chat = _reflection.GeneratedProtocolMessageType('Chat', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_CHAT,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.Chat)
    ))
  ,

  ProtocolMessage = _reflection.GeneratedProtocolMessageType('ProtocolMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_PROTOCOLMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.ProtocolMessage)
    ))
  ,

  ContactsArrayMessage = _reflection.GeneratedProtocolMessageType('ContactsArrayMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_CONTACTSARRAYMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.ContactsArrayMessage)
    ))
  ,

  HighlyStructuredMessage = _reflection.GeneratedProtocolMessageType('HighlyStructuredMessage', (_message.Message,), dict(
    DESCRIPTOR = _MESSAGE_HIGHLYSTRUCTUREDMESSAGE,
    __module__ = 'e2e_pb2'
    # @@protoc_insertion_point(class_scope:Message.HighlyStructuredMessage)
    ))
  ,
  DESCRIPTOR = _MESSAGE,
  __module__ = 'e2e_pb2'
  # @@protoc_insertion_point(class_scope:Message)
  ))
_sym_db.RegisterMessage(Message)
_sym_db.RegisterMessage(Message.SenderKeyDistributionMessage)
_sym_db.RegisterMessage(Message.ImageMessage)
_sym_db.RegisterMessage(Message.ContactMessage)
_sym_db.RegisterMessage(Message.LocationMessage)
_sym_db.RegisterMessage(Message.ExtendedTextMessage)
_sym_db.RegisterMessage(Message.DocumentMessage)
_sym_db.RegisterMessage(Message.AudioMessage)
_sym_db.RegisterMessage(Message.VideoMessage)
_sym_db.RegisterMessage(Message.Call)
_sym_db.RegisterMessage(Message.Chat)
_sym_db.RegisterMessage(Message.ProtocolMessage)
_sym_db.RegisterMessage(Message.ContactsArrayMessage)
_sym_db.RegisterMessage(Message.HighlyStructuredMessage)


# @@protoc_insertion_point(module_scope)



from yowsup.layers.protocol_messages.protocolentities import MessageProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.axolotl.protocolentities.enc import EncProtocolEntity
class EncryptedMessageProtocolEntity(MessageProtocolEntity):
    '''
    <message retry="1" from="49xxxxxxxx@s.whatsapp.net" t="1418906418" offline="1" type="text" id="1418906377-1" notify="Tarek Galal">
<enc type="{{type}}" mediatype="image|audio|location|document|contact" v="{{1 || 2}}">
HEX:33089eb3c90312210510e0196be72fe65913c6a84e75a54f40a3ee290574d6a23f408df990e718da761a210521f1a3f3d5cb87fde19fadf618d3001b64941715efd3e0f36bba48c23b08c82f2242330a21059b0ce2c4720ec79719ba862ee3cda6d6332746d05689af13aabf43ea1c8d747f100018002210d31cd6ebea79e441c4935f72398c772e2ee21447eb675cfa28b99de8d2013000</enc>

</message>
    '''

    def __init__(self, encEntities, _type, messageAttributes):
        super(EncryptedMessageProtocolEntity, self).__init__(_type, messageAttributes)
        self.setEncEntities(encEntities)

    def setEncEntities(self, encEntities = None):
        assert type(encEntities) is list and len(encEntities) \
            , "Must have at least a list of minumum 1 enc entity"
        self.encEntities = encEntities

    def getEnc(self, encType):
        for enc in self.encEntities:
            if enc.type == encType:
                return enc

    def getEncEntities(self):
        return self.encEntities

    def toProtocolTreeNode(self):
        node = super(EncryptedMessageProtocolEntity, self).toProtocolTreeNode()
        participantsNode = ProtocolTreeNode("participants")
        for enc in self.encEntities:
            encNode = enc.toProtocolTreeNode()
            if encNode.tag == "to":
                participantsNode.addChild(encNode)
            else:
                node.addChild(encNode)

        if len(participantsNode.getAllChildren()):
            node.addChild(participantsNode)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = MessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = EncryptedMessageProtocolEntity
        entity.setEncEntities(
            [EncProtocolEntity.fromProtocolTreeNode(encNode) for encNode in node.getAllChildren("enc")]
        )
        return entity



from yowsup.layers.protocol_media.protocolentities.test_message_media import MediaMessageProtocolEntityTest
from yowsup.layers.protocol_media.protocolentities import ContactMediaMessageProtocolEntity
from yowsup.layers.protocol_messages.proto.e2e_pb2 import Message


class ContactMediaMessageProtocolEntityTest(MediaMessageProtocolEntityTest):
    def setUp(self):
        super(ContactMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = ContactMediaMessageProtocolEntity
        m = Message()
        contact_message = Message.ContactMessage()
        contact_message.display_name = "abc"
        contact_message.vcard = b"VCARD_DATA"
        m.contact_message.MergeFrom(contact_message)
        proto_node = self.node.getChild("proto")
        proto_node["mediatype"] = "contact"
        proto_node.setData(m.SerializeToString())



from .waresponseparser import ResponseParser
from yowsup.env import YowsupEnv

import sys
import logging
from axolotl.ecc.curve import Curve
from axolotl.ecc.ec import ECPublicKey
from yowsup.common.tools import WATools
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from yowsup.axolotl.factory import AxolotlManagerFactory
import struct
import random
import base64

if sys.version_info < (3, 0):
    import httplib
    from urllib import quote as urllib_quote

    if sys.version_info >= (2, 7, 9):
        # see https://github.com/tgalal/yowsup/issues/677
        import ssl

        ssl._create_default_https_context = ssl._create_unverified_context

else:
    from http import client as httplib
    from urllib.parse import quote as urllib_quote

logger = logging.getLogger(__name__)


class WARequest(object):
    OK = 200
    ENC_PUBKEY = Curve.decodePoint(
        bytearray([
            5, 142, 140, 15, 116, 195, 235, 197, 215, 166, 134, 92, 108,
            60, 132, 56, 86, 176, 97, 33, 204, 232, 234, 119, 77, 34, 251,
            111, 18, 37, 18, 48, 45
        ])
    )

    def __init__(self, config):
        """
       :type method: str
       :param config:
       :type config: yowsup.config.v1.config.Config
       """

        self.pvars = []
        self.port = 443
        self.type = "GET"
        self.parser = None
        self.params = []
        self.headers = {}

        self.sent = False
        self.response = None

        self._config = config
        self._p_in = str(config.phone)[len(str(config.cc)):]
        self._axolotlmanager = AxolotlManagerFactory() \
            .get_manager(self._config.phone)  # type: yowsup.axolotl.manager.Axolotlmanager

        if config.expid is None:
            config.expid = WATools.generateDeviceId()

        if config.fdid is None:
            config.fdid = WATools.generatePhoneId()

        if config.client_static_keypair is None:
            config.client_static_keypair = WATools.generateKeyPair()

        self.addParam("cc", config.cc)
        self.addParam("in", self._p_in)
        self.addParam("lg", "en")
        self.addParam("lc", "GB")
        self.addParam("mistyped", "6")
        self.addParam("authkey", self.b64encode(config.client_static_keypair.public.data))
        self.addParam("e_regid", self.b64encode(struct.pack('>I', self._axolotlmanager.registration_id)))
        self.addParam("e_keytype", self.b64encode(b"\x05"))
        self.addParam("e_ident", self.b64encode(self._axolotlmanager.identity.publicKey.serialize()[1:]))

        signedprekey = self._axolotlmanager.load_latest_signed_prekey(generate=True)
        self.addParam("e_skey_id", self.b64encode(struct.pack('>I', signedprekey.getId())[1:]))
        self.addParam("e_skey_val", self.b64encode(signedprekey.getKeyPair().publicKey.serialize()[1:]))
        self.addParam("e_skey_sig", self.b64encode(signedprekey.getSignature()))

        self.addParam("fdid", config.fdid)
        self.addParam("expid", self.b64encode(config.expid))

        self.addParam("network_radio_type", "1")
        self.addParam("simnum", "1")
        self.addParam("hasinrc", "1")
        self.addParam("pid", int(random.uniform(100, 9999)))
        self.addParam("rc", 0)
        if self._config.id:
            self.addParam("id", self._config.id)

    def setParsableVariables(self, pvars):
        self.pvars = pvars

    def onResponse(self, name, value):
        if name == "status":
            self.status = value
        elif name == "result":
            self.result = value

    def addParam(self, name, value):
        self.params.append((name, value))

    def removeParam(self, name):
        for i in range(0, len(self.params)):
            if self.params[i][0] == name:
                del self.params[i]

    def addHeaderField(self, name, value):
        self.headers[name] = value

    def clearParams(self):
        self.params = []

    def getUserAgent(self):
        return YowsupEnv.getCurrent().getUserAgent()

    def send(self, parser=None, encrypt=True, preview=False):
        logger.debug("send(parser=%s, encrypt=%s, preview=%s)" % (
            None if parser is None else "[omitted]",
            encrypt, preview
        ))
        if self.type == "POST":
            return self.sendPostRequest(parser)

        return self.sendGetRequest(parser, encrypt, preview=preview)

    def setParser(self, parser):
        if isinstance(parser, ResponseParser):
            self.parser = parser
        else:
            logger.error("Invalid parser")

    def getConnectionParameters(self):

        if not self.url:
            return "", "", self.port

        try:
            url = self.url.split("://", 1)
            url = url[0] if len(url) == 1 else url[1]

            host, path = url.split('/', 1)
        except ValueError:
            host = url
            path = ""

        path = "/" + path

        return host, self.port, path

    def encryptParams(self, params, key):
        """
        :param params:
        :type params: list
        :param key:
        :type key: ECPublicKey
        :return:
        :rtype: list
        """
        keypair = Curve.generateKeyPair()
        encodedparams = self.urlencodeParams(params)

        cipher = AESGCM(Curve.calculateAgreement(key, keypair.privateKey))
        ciphertext = cipher.encrypt(b'\x00\x00\x00\x00' + struct.pack('>Q', 0), encodedparams.encode(), b'')

        payload = base64.b64encode(keypair.publicKey.serialize()[1:] + ciphertext)
        return [('ENC', payload)]

    def sendGetRequest(self, parser=None, encrypt_params=True, preview=False):
        logger.debug("sendGetRequest(parser=%s, encrypt_params=%s, preview=%s)" % (
            None if parser is None else "[omitted]",
            encrypt_params, preview
        ))
        self.response = None

        if encrypt_params:
            logger.debug("Encrypting parameters")
            if logger.level <= logging.DEBUG:
                logger.debug("pre-encrypt (encoded) parameters = \n%s", (self.urlencodeParams(self.params)))
            params = self.encryptParams(self.params, self.ENC_PUBKEY)
        else:
            ## params will be logged right before sending
            params = self.params

        parser = parser or self.parser or ResponseParser()

        headers = dict(
            list(
                {
                    "User-Agent": self.getUserAgent(),
                    "Accept": parser.getMeta()
                }.items()
            ) + list(self.headers.items()))

        host, port, path = self.getConnectionParameters()

        self.response = WARequest.sendRequest(host, port, path, headers, params, "GET", preview=preview)

        if preview:
            logger.info("Preview request, skip response handling and return None")
            return None

        if not self.response.status == WARequest.OK:
            logger.error("Request not success, status was %s" % self.response.status)
            return {}

        data = self.response.read()
        logger.info(data)

        self.sent = True
        return parser.parse(data.decode(), self.pvars)

    def sendPostRequest(self, parser=None):
        self.response = None
        params = self.params  # [param.items()[0] for param in self.params];

        parser = parser or self.parser or ResponseParser()

        headers = dict(list({"User-Agent": self.getUserAgent(),
                             "Accept": parser.getMeta(),
                             "Content-Type": "application/x-www-form-urlencoded"
                             }.items()) + list(self.headers.items()))

        host, port, path = self.getConnectionParameters()
        self.response = WARequest.sendRequest(host, port, path, headers, params, "POST")

        if not self.response.status == WARequest.OK:
            logger.error("Request not success, status was %s" % self.response.status)
            return {}

        data = self.response.read()

        logger.info(data)

        self.sent = True
        return parser.parse(data.decode(), self.pvars)

    def b64encode(self, value):
        return base64.urlsafe_b64encode(value).replace(b'=', b'')

    @classmethod
    def urlencode(cls, value):
        if type(value) not in (str, bytes):
            value = str(value)

        out = ""
        for char in value:
            if type(char) is int:
                char = bytearray([char])
            quoted = urllib_quote(char, safe='')
            out += quoted if quoted[0] != '%' else quoted.lower()

        return out \
            .replace('-', '%2d') \
            .replace('_', '%5f') \
            .replace('~', '%7e')

    @classmethod
    def urlencodeParams(cls, params):
        merged = []
        for k, v in params:
            merged.append(
                "%s=%s" % (k, cls.urlencode(v))
            )
        return "&".join(merged)

    @classmethod
    def sendRequest(cls, host, port, path, headers, params, reqType="GET", preview=False):
        logger.debug("sendRequest(host=%s, port=%s, path=%s, headers=%s, params=%s, reqType=%s, preview=%s)" % (
            host, port, path, headers, params, reqType, preview
        ))

        params = cls.urlencodeParams(params)

        path = path + "?" + params if reqType == "GET" and params else path

        if not preview:
            logger.debug("Opening connection to %s" % host)
            conn = httplib.HTTPSConnection(host, port) if port == 443 else httplib.HTTPConnection(host, port)
        else:
            logger.debug("Should open connection to %s, but this is a preview" % host)
            conn = None

        if not preview:
            logger.debug("Sending %s request to %s" % (reqType, path))
            conn.request(reqType, path, params, headers)
        else:
            logger.debug("Should send %s request to %s, but this is a preview" % (reqType, path))
            return None

        response = conn.getresponse()
        return response
